"""
Define classes for handling scenarios


This module implements the classes used to represent the four
main types of scenarios in Atomica:

- Parameter scenarios
- Budget scenarios
- Capacity scenarios
- Coverage scenarios

There are broadly two kinds of scenario

- Parameter scenarios which involve modifying a :class:`ParameterSet` instance
- Program scenarios (budget, capacity, coverage) which involve modifying a :class:`ProgramInstructions` instance

"""

import numpy as np
import sciris as sc
from .system import logger
from .utils import NamedItem
from .programs import ProgramInstructions, ProgramSet
from .parameters import ParameterSet
from .results import Result

class Scenario(NamedItem):
    """
    Base scenario class

    All Scenario objects share this type. It is a NamedItem that also has an
    ``active`` property which represents whether the scenario should be run
    as part of :meth:`project.Project.run_scenarios`

    The optional variables ``parsetname`` and ``progsetname`` reference the name of parsets
    and progsets to use via :meth:`Scenario.run`. These should match the names of objects in the
    project passed into :meth:`Scenario.run`. However, :meth:`Scenario.run` can also take in
    a parset and/or progset directly, allowing the scenario to be used with other parsets and
    progsets on the fly. If  ``parsetname`` and ``progsetname`` are not set in the :class:`Scenario`
    then they must be passed into :meth:`Scenario.run`.

    :param name: The name of the scenario - also sets the result name via :meth:`Scenario.run`
    :param parsetname: If running via ``Project.run_scenarios`` this identifies which parset to use from the project
    :param progsetname: If running via ``Project.run_scenarios`` this identifies which progset to use. If set to ``None`` then programs will not be used
    :param active: If running via ``Project.run_scenarios`` this flags whether to run the scenario

    """

    def __init__(self, name:str, active:bool=True, parsetname:str =None, progsetname:str =None):
        NamedItem.__init__(self, name)
        self.parsetname = parsetname
        self.progsetname = progsetname
        self.active = active

    def get_parset(self,parset,project) -> ParameterSet:
        """
        Get scenario parset

        If the derived scenario class modifies the parset, return the modified version

        :param parset: Input :class:`ParameterSet`
        :return: Modified parset for use in the simulation

        """

        return parset

    def get_progset(self, progset:ProgramSet, project) -> ProgramSet:
        """
        Get scenario progset

        If the derived scenario class modifies the progset, return the modified version

        :param progset: Input :class:`ProgramSet`
        :return: Modified progset for use in the simulation

        """

        return progset

    def get_instructions(self, progset:ProgramSet, project) -> ProgramInstructions:

        """
        Get scenario instructions

        If the derived scenario class produces  program instructions, return
        them here.

        :param progset: Input :class:`ProgramSet`
        :return: :class:`ProgramInstructions` instance, or None if no instructions (in which case, programs will not be used)

        """

        return None

    def run(self, project, parset:ParameterSet =None, progset:ProgramSet =None, store_results:bool =True) -> Result:
        """
        Run scenario

        :param project: A :class:`Project` instance
        :param parset: Optionally a :class:`ParameterSet` instance, otherwise will use ``self.parsetname``
        :param progset: Optionally a :class:`ProgramSet` instance, otherwise will use ``self.progsetname``
        :param store_results: If True, the results will be copied into the project
        :return: A :class:`Result` object

        """

        if parset is None:
            parset = project.parsets[self.parsetname]
        else:
            parset = project.parset(parset)

        if progset is None and self.progsetname is not None:
            progset = project.progsets[self.progsetname]
        elif progset is not None:
            progset = project.progset(progset)

        parset = self.get_parset(parset, project)
        progset = self.get_progset(progset, project)
        instructions = self.get_instructions(progset, project)

        if progset is not None:
            if instructions is None:
                raise Exception('If using programs, the scenario must contain instructions specifying at minimum the program start year')
            result = project.run_sim(parset=parset, progset=progset, progset_instructions=instructions, result_name=self.name, store_results=store_results)
        else:
            result = project.run_sim(parset=parset, result_name=self.name, store_results=store_results)

        return result


class CombinedScenario(Scenario):
    """
    Define combined (budget+program) scenario

    This object stores both a set of scenario values and a set of program instructions.
    This allows it to simultaneously apply parameter, budget, and coverage overwrites.

    As usual, parameter values from programs take precedence over parameter values from parsets, and
    within programs, coverage takes precedence ovetoxr budgets.

    :param name: The name of the scenario. This will also be used to name the result
    :param active: If running via ``Project.run_scenarios`` this flags whether to run the scenario
    :param parsetname: If running via ``Project.run_scenarios`` this identifies which parset to use from the project
    :param progsetname: If running via ``Project.run_scenarios`` this identifies which progset to use. If set to ``None`` then programs will not be used
    :param scenario_values: Parameter value overwrites, used as input to :class:`ParameterScenario`
    :param instructions: A :class`ProgramInstructions` instance containing required program overwrites (budget, capacity, coverage)

    """

    def __init__(self, name:str =None, active:bool =True, parsetname:str =None, progsetname:str =None, scenario_values:dict =None, instructions:ProgramInstructions =None):
        super().__init__(name, active, parsetname, progsetname)
        self.scenario_values = scenario_values
        self.instructions = instructions

    def get_parset(self, parset, project) -> ParameterSet:
        if self.scenario_values is not None:
            scenario_parset = ParameterScenario(scenario_values=self.scenario_values).get_parset(parset, project)
        else:
            scenario_parset = parset
        return scenario_parset

    def get_instructions(self, progset:ProgramSet, project) -> ProgramInstructions:
        return self.instructions


class BudgetScenario(Scenario):

    def __init__(self, name=None, active:bool=True, parsetname:str =None, progsetname:str =None, alloc:dict =None, start_year=2018):
        super().__init__(name, active, parsetname, progsetname)
        self.start_year = start_year # Program start year
        self.alloc = sc.dcp(alloc) if alloc is not None else sc.odict()

    def get_instructions(self, progset:ProgramSet, project) -> ProgramInstructions:
        return ProgramInstructions(start_year=self.start_year, alloc=self.alloc)


class CoverageScenario(Scenario):

    def __init__(self, name=None, active:bool=True, parsetname:str =None, progsetname:str =None, coverage:dict =None, start_year=2018):
        super().__init__(name, active, parsetname, progsetname)
        self.start_year = start_year # Program start year
        self.coverage = sc.dcp(coverage) if coverage is not None else sc.odict()

    def get_instructions(self, progset:ProgramSet, project) -> ProgramInstructions:
        return ProgramInstructions(start_year=self.start_year, coverage=self.coverage)


class ParameterScenario(Scenario):
    def __init__(self, name:str =None, scenario_values:dict =None, active:bool=True, parsetname:str =None):
        """
        Define and run parameter scenarios

        This object stores overwrites to parameter values that are used to modify
        a :class:`ParameterSet` instance before running a simulation.

        :param name: The name of the scenario. This will also be used to name the result
        :param scenario_values: A dict of overwrites to parameter values. The structure is
            ``{parameter_label: {pop_label: dict o}`` where the overwrite ``o`` contains keys
             - ``t`` : np.array or list with year values
             - ``y`` : np.array or list with corresponding parameter values
             - ``smooth_onset`` (optional): Smoothly ramp parameter value rather than having a stepped change
        :param active: If running via ``Project.run_scenarios`` this flags whether to run the scenario
        :param parsetname: If running via ``Project.run_scenarios`` this identifies which parset to use from the project

        Example usage:

        >>> scvalues = dict()
        >>> param = 'birth_transit'
        >>> scvalues[param] = dict()
        >>> scvalues[param]['Pop1'] = dict()
        >>> scvalues[param]['Pop1']['y'] = [3e6, 1e4, 1e4, 2e6]
        >>> scvalues[param]['Pop1']['t'] = [2003.,2004.,2014.,2015.]
        >>> scvalues[param]['Pop1']['smooth_onset'] = 1
        >>> scvalues[param]['Pop1']['smooth_onset'] = [1,2,3,4] (same length as y)
        >>> pscenario = ParameterScenario(name="examplePS",scenario_values=scvalues)

        """
        super().__init__(name, active, parsetname)
        self.parsetname = parsetname
        # TODO - could do some extra validation here
        self.scenario_values = sc.dcp(scenario_values) if scenario_values is not None else dict()

    def get_parset(self, parset:ParameterSet, project) -> ParameterSet:
        """
        Return modified parset

        This method takes in a :class:`ParameterSet` and modifies it by applying the overwrites
        present in the scenario. This can thus be used to return the :class:`ParameterSet` for use in
        other simulations that are manually run, or to do things like perform a budget scenario simulation
        in conjunction with a parameter scenario.

        Get the corresponding parameterSet for this scenario, given an input parameterSet for the default baseline
        activity.

        The output depends on whether to overwrite (replace) or add values that appear in both the
        parameterScenario's parameterSet to the baseline parameterSet.

        :param parset: A :class:`ParameterSet` instance
        :param settings: A :class:`ProjectSettings` instance (e.g. ``proj.settings``)
        :return: A new, modified :class:`ParameterSet` object

        """

        # Note - the parset will be overwritten between the first and last year specified in scvalues
        # on a per-parameter+pop basis. Within the scenario, only the data points in scvalues will be used

        new_parset = sc.dcp(parset)
        new_parset.name = self.name + '_' + parset.name

        for par_label in self.scenario_values.keys():
            par = new_parset.pars[par_label]  # This is the parameter we are updating
            has_function = project.framework.pars.at[par.name, 'function'] # Flag whether this is a function parameter in the framework

            for pop_label, overwrite in self.scenario_values[par_label].items():

                # Remove Nones and Nans
                overwrite = sc.dcp(overwrite)
                overwrite['t'] = sc.promotetoarray(overwrite['t']).astype('float') # astype('float') converts None to np.nan
                overwrite['y'] = sc.promotetoarray(overwrite['y']).astype('float')
                idx = ~np.isnan(overwrite['t']) & ~np.isnan(overwrite['y'])
                if not np.any(idx):
                    continue
                overwrite['t'] = overwrite['t'][idx]
                overwrite['y'] = overwrite['y'][idx]

                original_y_end = par.interpolate(np.array([max(overwrite['t']) + 1e-5]), pop_label)

                # If the Parameter had an assumption, then insert the assumption value in the start year
                if not par.ts[pop_label].has_time_data:
                    par.ts[pop_label].insert(project.settings.sim_start, par.ts[pop_label].assumption)

                if has_function and 'smooth_onset' in overwrite:
                    raise Exception('Parameter function overwrites cannot have smooth onsets (because the value at the onset time is not yet known)')

                if 'smooth_onset' not in overwrite:
                    # Note parameter functions still get smooth onset set here - this ensures
                    # correct non-smooth-onset behaviour _during_ the overwrite
                    overwrite['smooth_onset'] = 1e-5

                if np.isscalar(overwrite['smooth_onset']):
                    onset = [overwrite['smooth_onset']]*len(overwrite['y'])
                else:
                    assert len(overwrite['smooth_onset']) == len(overwrite['y']), 'Smooth onset must be either a scalar or an array with length matching y-values'
                    onset = overwrite['smooth_onset']

                # Now, insert all of the program overwrites
                if len(overwrite['t']) != len(overwrite['y']):
                    raise Exception('Number of time points in overwrite does not match number of values')

                if len(overwrite['t']) == 1:
                    raise Exception('Only one time point was specified in the overwrite, which means that the overwrite will not have any effect')

                for i in range(0, len(overwrite['t'])):

                    # Account for smooth onset
                    if onset[i] > 0:
                        t = overwrite['t'][i] - onset[i]

                        if i > 0 and t <= overwrite['t'][i - 1]:
                            # If the smooth onset extends to before the previous point, then just use the
                            # previous point directly instead
                            y = overwrite['y'][i - 1] / par.y_factor[pop_label] / par.meta_y_factor
                            par.ts[pop_label].remove_between([overwrite['t'][i - 1], overwrite['t'][i]])
                            par.ts[pop_label].insert(t,y)
                        else:
                            # Otherwise, get the value at the smooth onset time, add it as a control
                            # point, and remove any intermediate points
                            y = par.interpolate(np.array([t]), pop_label)
                            par.ts[pop_label].remove_between([t, overwrite['t'][i]])  # Remove values during onset period
                            par.ts[pop_label].insert(t,y)
                    elif i > 0:
                        # If not doing smooth onset, and this is not the first point being overwritten,
                        # then remove all control points between this point and the last one
                        par.ts[pop_label].remove_between([overwrite['t'][i - 1], overwrite['t'][i]])

                    # Insert the overwrite value - assume scenario value is AFTER y-factor rescaling
                    par.ts[pop_label].insert(overwrite['t'][i], overwrite['y'][i] / par.y_factor[pop_label] / par.meta_y_factor)

                # Add an extra point to return the parset back to it's original value after the final overwrite
                par.ts[pop_label].insert(max(overwrite['t']) + 1e-5, original_y_end)

                if has_function:
                    par.skip_function[pop_label] = (min(overwrite['t']), max(overwrite['t']))

        return new_parset




"""
Implements various Optimizations in Atomica

This module implements the :class:`Optimization` class, which contains the
information required to perform an optimization in Atomica. An Optimization
effectively serves as a mapping from one set of program instructions to another.

"""

import sciris as sc
from .system import logger, NotFoundError
from .utils import NamedItem
import numpy as np
from .model import Model, Link
import pickle
import scipy.optimize
from collections import defaultdict
from .utils import TimeSeries
from .results import Result
from .cascade import get_cascade_vals
from .programs import ProgramInstructions


class InvalidInitialConditions(Exception):
    # This error gets thrown if the initial conditions yield an objective value
    # that is not finite
    pass


class UnresolvableConstraint(Exception):
    # This error gets thrown if it is _impossible_ to satisfy the constraints. There are
    # two modes of constraint failure
    # - The constraint might not be satisfied on this iteration, but could be satisfied by other
    #   parameter values
    # - The constraint is impossible to satisfy because it is inconsistent (for example, if the
    #   total spend is greater than the sum of the upper bounds on all the individual programs)
    #   in which case the algorithm cannot continue
    # This error gets raised in the latter case, while the former should result in the iteration
    # being skipped
    pass


class Adjustable(object):

    def __init__(self, name, limit_type='abs', lower_bound=-np.inf, upper_bound=np.inf, initial_value=None):
        self.name = name  # identify the adjustable
        self.limit_type = limit_type  # 'abs' or 'rel'
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.initial_value = initial_value  # This could be None, but if it is None, then `extract_from_instructions` cannot be None

    def get_hard_bounds(self, x0=None):
        # Return hard bounds based the limit type and specified bounds
        self.lower_bound = self.lower_bound if self.lower_bound is not None else -np.inf
        self.upper_bound = self.upper_bound if self.upper_bound is not None else np.inf
        xmin = self.lower_bound if self.limit_type == 'abs' else x0 * self.lower_bound
        xmax = self.upper_bound if self.limit_type == 'abs' else x0 * self.upper_bound
        return xmin, xmax


class Adjustment(object):
    def __init__(self, name):
        self.name = name
        self.adjustables = None

    def get_initialization(self, progset, instructions):
        # Return initial values for ASD
        return [x.initial_value for x in self.adjustables]

    def update_instructions(self, adjustable_values, instructions):
        # adjustable_values contains the values for each adjustable in self.adjustables
        # at the current ASD iteration. This function updates the provided instructions in place
        return


class SpendingAdjustment(Adjustment):
    # This adjustment class represents making a spending quantity adjustable. By default, the
    # base class simply overwrites the spending value at a particular point in time
    # A SpendingAdjustment has a separate Adjustable for each time reached (independently)
    def __init__(self, prog_name, t, limit_type='abs', lower=0.0, upper=np.inf, initial=None):
        Adjustment.__init__(self, name=prog_name)
        self.prog_name = prog_name
        self.t = sc.promotetoarray(t)  # Time at which to apply the adjustment

        lower = sc.promotetolist(lower,keepnone=True)
        if len(lower) == 1:
            lower = lower*len(self.t)
        else:
            assert len(lower) == len(self.t), "If supplying lower bounds, you must either specify one, or one for every time point"

        upper = sc.promotetolist(upper,keepnone=True)
        if len(upper) == 1:
            upper = upper*len(self.t)
        else:
            assert len(upper) == len(self.t), "If supplying upper bounds, you must either specify one, or one for every time point"

        initial = sc.promotetolist(initial,keepnone=True)
        if len(initial) == 1:
            initial = initial*len(self.t)
        else:
            assert len(initial) == len(self.t), "If supplying initial values, you must either specify one, or one for every time point"

        self.adjustables = [Adjustable(prog_name, limit_type, lower_bound=lb, upper_bound=ub, initial_value=init) for lb,ub,init in zip(lower,upper,initial)]

    def update_instructions(self, adjustable_values, instructions):
        # There is one Adjustable for each time point, so the adjustable_values
        # are a list of this same length, one value for each time point
        for i, t in enumerate(self.t):
            if self.prog_name not in instructions.alloc:
                instructions.alloc[self.prog_name] = TimeSeries(t=t, vals=adjustable_values[i])
            else:
                instructions.alloc[self.prog_name].insert(t, adjustable_values[i])

    def get_initialization(self, progset, instructions):
        initialization = []
        for adjustable, t in zip(self.adjustables, self.t):
            if adjustable.initial_value:
                initialization.append(adjustable.initial_value)
            else:
                alloc = progset.get_alloc(t, instructions)
                initialization.append(alloc[self.prog_name][0])  # The Adjustable's name corresponds to the name of the program being overwritten.
        return initialization


class StartTimeAdjustment(Adjustment):
    # Example of an Adjustment that does not target a spending value

    def __init__(self, name, lower, upper, initial):
        Adjustment.__init__(self, name=name)
        self.adjustables = [Adjustable('start_year', limit_type='abs', lower=lower, upper=upper, initial=initial)]

    def update_instructions(self, adjustable_values, instructions):
        instructions.start_year = adjustable_values[0]

    def get_initialization(self, progset, instructions):
        if self.initial_value:
            return self.initial_value
        else:
            return instructions.start_year


class ExponentialSpendingAdjustment(Adjustment):
    # Example of a parametric time-varying budget where multiple parameters
    # are mapped to spending via a function

    def __init__(self, prog_name, t, t_0, t_end, p1, a1, a2):
        Adjustment.__init__(self, name=prog_name)
        self.prog_name = prog_name
        self.t = t  # Vector of time values instructions are updated at
        self.t_0 = t_0  # non-adjustable parameter
        self.t_end = t_end  # non_adjustable parameter
        self.adjustables = [Adjustable('p1', initial_value=p1), Adjustable('a1', initial_value=a1), Adjustable('a2', initial_value=a2)]  # Would need to specify initial values and limits for these parameters

    # Note - we can use get_initialization from the base class because all of the Adjustables
    # have explicit initial values. Note that it's essential to provide explicit initial values
    # for any Adjustable that does not explicitly appear in the instructions. Although in theory,
    # it would be possible to write a `get_initialization` function that fits the Adjustables
    # to the initial spending...

    def update_instructions(self, adjustable_values, instructions):
        # p1*exp(a1*(t-t_0))*exp(b1*(t-t_end))
        instructions.alloc[self.prog_name][self.t] = adjustable_values[0] * np.exp(adjustable_values[1] * (self.t - self.t_0)) * np.exp(adjustable_values[2] * (self.t - self.t_end))


class PairedLinearSpendingAdjustment(Adjustment):
    # This example shows a parametric time-varying budget reaching more than one program.
    # A single adjustable corresponding to the rate of change simultaneously acts on
    # two programs in opposite directions

    def __init__(self, prog_names, t):
        Adjustment.__init__(self, name=None)
        assert len(prog_names) == 2, 'PairedLinearSpendingAdjustment needs exactly two program names'
        self.prog_names = prog_names
        self.t = t  # [t_start,t_stop] for when to start/stop ramping
        self.adjustables = [Adjustable('ramp', initial_value=0.0)]

    def update_instructions(self, adjustable_values, instructions):

        gradient = adjustable_values[0]
        tspan = (self.t[1] - self.t[0])

        if gradient < 0:  # We are transferring money from program 2 to program 1
            available = -instructions.alloc[self.prog_names[1]].get(self.t[0])
            max_gradient = float(available) / tspan
            if gradient < max_gradient:
                gradient = max_gradient
        else:
            available = instructions.alloc[self.prog_names[0]].get(self.t[0])
            max_gradient = float(available) / tspan
            if gradient > max_gradient:
                gradient = max_gradient

        funding_change = gradient * tspan  # Amount transferred from program 1 to program 2

        # This does not change the amount of funds allocated in the initial year
        for prog in self.prog_names:
            instructions.alloc[prog].insert(self.t[0], instructions.alloc[prog].interpolate(self.t[0])[0])
            instructions.alloc[prog].remove_between(self.t)

        instructions.alloc[self.prog_names[0]].insert(self.t[1], instructions.alloc[self.prog_names[0]].get(self.t[0]) - funding_change)
        instructions.alloc[self.prog_names[1]].insert(self.t[1], instructions.alloc[self.prog_names[1]].get(self.t[0]) + funding_change)


class Measurable(object):

    def __init__(self, measurable_name, t, pop_names=None, weight=1.0, fcn=None):
        self.measurable_name = measurable_name
        self.t = sc.promotetoarray(t)  # Single year, or multiple years
        self.weight = weight
        self.pop_names = pop_names
        self.fcn = fcn  # transformation function to apply

    def eval(self, model):
        # This is the main interface with the optimization code - this function gets called
        # to return the transformed, weighted objective value from this Measurable for use in ASD
        # Only overload this if you want to customize the transformation and weighting
        val = self.get_objective_val(model)
        return self._transform_val(val)

    def get_objective_val(self, model):
        # This returns the base objective value, prior to any function transformation
        # or weighting. The function transformation and weighting are handled by this base
        # class - derived classes may wish to not expose things like the function mapping
        # in their constructors, if that wouldn't be appropriate. But otherwise, they can
        # inherit this behaviour for free. So derived classes should overload this method
        #
        # The base class has the default behaviour that the 'measurable name' is a model variable
        if len(self.t) == 1:
            t_filter = model.t == self.t  # boolean vector for whether to use the time point or not
        else:
            t_filter = (model.t >= self.t[0]) & (model.t < self.t[1])  # Don't include upper bound, so [2018,2019] will include exactly one year

        if self.measurable_name in model.progset.programs:
            alloc = model.progset.get_alloc(model.t, model.program_instructions)
            val = np.sum(alloc[self.measurable_name][t_filter])
        else:  # If the measurable is a model output...
            val = 0.0
            matched = False # Flag whether any variables were found
            for pop in model.pops:
                if not self.pop_names:
                    # If no pops were provided, then iterate over all pops but skip those where the measureable is not defined
                    # Use this approach rather than checking the pop type in the framework because user could be optimizating
                    # flow rates or transitions that don't appear in the framework
                    try:
                        vars = pop.get_variable(self.measurable_name)
                        matched = True
                    except NotFoundError:
                        continue
                elif pop not in self.pop_names:
                    continue
                else:
                    vars = pop.get_variable(self.measurable_name) # If variable is missing and the pop was explicitly defined, raise the error
                    matched = True

                for var in vars:
                    if isinstance(var, Link):
                        val += np.sum(var.vals[t_filter] / var.dt)  # Annualize link values - usually this won't make a difference, but it matters if the user mixes Links with something else in the objective
                    else:
                        val += np.sum(var.vals[t_filter])

            if not matched:
                # Raise an error if the measureable was not found in any populations
                raise Exception('"%s" not found in any populations' % (self.measurable_name))

        return val

    def _transform_val(self, val):
        # Apply function transformation and weight to the value
        if self.fcn:
            val = self.fcn(val)
        val *= self.weight
        return val


class MinimizeMeasurable(Measurable):
    # Syntactic sugar for a measurable that minimizes its quantity
    def __init__(self, measurable_name, t, pop_names=None):
        Measurable.__init__(self, measurable_name, t=t, weight=1, pop_names=pop_names)


class MaximizeMeasurable(Measurable):
    # Syntactic sugar for a measurable that maximizes its quantity
    def __init__(self, measurable_name, t, pop_names=None):
        Measurable.__init__(self, measurable_name, t=t, weight=-1, pop_names=pop_names)


class AtMostMeasurable(Measurable):
    # A Measurable that impose a penalty if the quantity is larger than some threshold
    # The initial points should be 'valid' in the sense that the quantity starts out
    # below the threshold (and during ASD it will never be allowed to cross the threshold)
    def __init__(self, measurable_name, t, threshold, pop_names=None):
        Measurable.__init__(self, measurable_name, t=t, weight=np.inf, pop_names=pop_names)
        self.weight = 1.0
        self.threshold = threshold
        self.fcn = lambda x: np.inf if x > self.threshold else 0.0

    def __repr__(self):
        return 'AtMostMeasurable(%s < %f)' % (self.measurable_name, self.threshold)


class AtLeastMeasurable(Measurable):
    # A Measurable that impose a penalty if the quantity is smaller than some threshold
    # The initial points should be 'valid' in the sense that the quantity starts out
    # above the threshold (and during ASD it will never be allowed to cross the threshold)
    def __init__(self, measurable_name, t, threshold, pop_names=None):
        Measurable.__init__(self, measurable_name, t=t, weight=np.inf, pop_names=pop_names)
        self.weight = 1.0
        self.threshold = threshold
        self.fcn = lambda x: np.inf if x < self.threshold else 0.0

    def __repr__(self):
        return 'AtLeastMeasurable(%s > %f)' % (self.measurable_name, self.threshold)


class MaximizeCascadeStage(Measurable):
    # This Measurable will maximize the number of people in a cascade stage, whatever that stage is
    # e.g. `measurable = MaximizeCascadeStage('main',2020)
    # If multiple years are provided, they will be summed over
    # Could add option to weight specific years later on, if desired

    def __init__(self, cascade_name, t, pop_names='all', weight=1.0, cascade_stage=-1):
        # Instantiate thpop_names can be a single pop name (including all), or a list of pop names
        # aggregations are supported by setting pop_names to a dict e.g.
        # pop_names = {'foo':['0-4','5-14']}
        #
        # The cascade name is passed to `get_cascade_vals` so any cascade can be passed in
        # Usually this would be 'None' (fallback cascade, or first cascade) or the name of a
        # cascade
        #
        # - t : Can be a single time or array of times (as supposed by get_cascade_vals)
        # - cascade_stage specifies which stage(s) to include. It can be
        #   - An integer indexing the `cascade_vals` dict e.g. -1 is the last stage
        #   - The name of a cascade stage
        #   - A list of integers and/or names to include multiple stages

        Measurable.__init__(self, cascade_name, t=t, weight=-weight, pop_names=pop_names)
        if not isinstance(cascade_stage, list):
            cascade_stage = [cascade_stage]
        self.cascade_stage = cascade_stage
        if not isinstance(self.pop_names, list):
            self.pop_names = [self.pop_names]

    def get_objective_val(self, model):
        result = Result(model=model)
        val = 0
        for pop_name in self.pop_names:
            cascade_vals = get_cascade_vals(result, self.measurable_name, pop_name, self.t)
            for stage in self.cascade_stage:  # Loop over included stages
                val += np.sum(cascade_vals[0][stage])  # Add the values from the stage, summed over time
        return val


class MaximizeCascadeConversionRate(Measurable):
    """
    Maximize overall conversion rate

    Maximize conversion summed over all cascade stages

    :param cascade_name: The name of one of the cascades in the Framework
    :param t: A single time value e.g. 2020
    :param pop_names: A single pop name (including 'all'), a list of populations,
                  or a dict/list of dicts, each with a single aggregation e.g. ``{'foo':['0-4','5-14']}``
    :param weight: Weighting factor for this Measurable in the overall objective function

    """

    def __init__(self, cascade_name, t:float, pop_names='all', weight=1.0):
        Measurable.__init__(self, cascade_name, t=t, weight=-weight, pop_names=pop_names)
        if not isinstance(self.pop_names, list):
            self.pop_names = [self.pop_names]

    def get_objective_val(self, model):
        if self.t < model.t[0] or self.t > model.t[-1]:
            raise Exception('Measurable year for optimization (%d) is outside the simulation range (%d-%d)' % (self.t,model.t[0],model.t[-1]))
        result = Result(model=model)
        val = 0
        for pop_name in self.pop_names:
            cascade_vals = get_cascade_vals(result, self.measurable_name, pop_name, self.t)[0]
            cascade_array = np.hstack(cascade_vals.values())
            conversion = cascade_array[1:] / cascade_array[0:-1]
            val += np.sum(conversion)
        return val


class Constraint(object):
    # A Constraint represents a condition that must be satisfied by the Instructions
    # after the cumulative effect of all adjustments. The Instructions are rescaled to
    # satisfy the constraint directly (rather than changing the value of the Adjustables)
    # although this distinction really only matters in the context of parametric spending

    def get_hard_constraint(self, optimization, instructions):
        return

    def constrain_instructions(self, instructions, hard_constraints):
        # Constrains the instructions, returns a metric penalizing the constraint
        # If there is no penalty associated with adjusting (perhaps if all of the Adjustments are
        # parametric?) then this would be 0.0
        return 0.0


class TotalSpendConstraint(Constraint):
    # This class implements a constraint on the total spend at every time point when a program
    # is optimizable. A program is considered optimizable if an Adjustment reaches that program
    # at the specified time. Spending is constrained independently at all times when any program
    # is adjustable.
    #
    # Important - this constraint only acts on program spending that is reached by an Adjustment.

    def __init__(self, total_spend=None, t=None, budget_factor=1.0):
        # total_spend allows the total spending in a particular year to be explicitly specified
        # rather than drawn from the initial allocation. This could be useful when using parametric
        # programs.
        # This constraint can be set to only apply in certain years.
        # The budget_factor multiplies the total spend at the time the hard_constraint is assigned
        # Typically this is to scale up the available spending when that spending is being drawn from
        # the instructions/progset (otherwise the budget_factor could already be part of the specified total spend)
        #
        # INPUTS
        # - t : A time, or list of times, at which to apply the total spending constraint. If None, it will
        #       automatically be set to all years in which spending adjustments are being made
        # - total_spend: A list of spending amounts the same size as t (can contain Nones), or None.
        #                For times in which the total spend is None, it will be automatically set to the sum of
        #                spending on optimizable programs in the corresponding year
        # - budget_factor: The budget factor multiplies whatever the total_spend is. This can either be a single value, or
        #                  a year specific value
        #
        # Note that if no times are specified, the budget factor should be a scalar but no explicit
        # spending values can be specified. This is because in the case where different programs are
        # optimized in different years, an explicit total spending constraint applying to all
        # times is unlikely to be a sensible choice (so we just ask the user to specify the time as well)
        self.total_spend = sc.promotetoarray(total_spend) if total_spend is not None else ()
        self.t = sc.promotetoarray(t) if t is not None else ()
        self.budget_factor = sc.promotetoarray(budget_factor)

        if t is None:
            assert total_spend is None, 'If no times are specified, no total spend values can be specified either'
            assert len(self.budget_factor) == 1, 'If no times are specified, the budget factor must be scalar'

        if t is not None and total_spend is not None:
            assert len(self.total_spend) == len(self.t), 'If specifying both the times and values for the total spending constraint, their lengths must be the same'
        if len(self.budget_factor) > 1:
            assert len(self.budget_factor) == len(self.t), 'If specifying multiple budget factors, you must also specify the years in which they are used'

    def get_hard_constraint(self, optimization, instructions):
        # First, at each time point where a program overwrite exists, we need to store
        # the names of all of the programs being overwritten
        # e.g.
        # hard_constraints['programs'][2020] = ['Program 1','Program 2']
        # hard_constraints['programs'][2030] = ['Program 2','Program 3']
        # By convention, an Adjustable affecting a program should store the program's name in the
        # `prog_name` attribute. If a program reaches multiple programs, then `prog_name` will be a list

        # First, we make a dictionary storing the years in which any adjustments are made
        # as well as the upper and lower bounds in that year. This is only done for spending
        # adjustments (i.e. those where `adjustment.prog_name` is defined)
        hard_constraints = {}
        hard_constraints['programs'] = defaultdict(set)  # It's a set so that it will work properly if multiple Adjustments reach the same parameter at the same time. However, this would a bad idea and nobody should do this!
        for adjustment in optimization.adjustments:
            if hasattr(adjustment, 'prog_name'):
                for t in list(adjustment.t):
                    if isinstance(adjustment.prog_name, list):
                        hard_constraints['programs'][t].update(adjustment.prog_name)
                    else:
                        hard_constraints['programs'][t].add(adjustment.prog_name)

        if len(self.t):
            # Check that every explictly specified time has
            # corresponding spending adjustments available for use
            missing_times = set(self.t) - set(hard_constraints['programs'].keys())
            if missing_times:
                raise Exception('Total spending constraint was specified in %s but the optimization does not have any adjustments at those times' % (missing_times))

        # Now we have a set of times and programs for which we need to get total spend, and
        # also which programs should be included in the total for that year
        #
        # hard_constraints['total_spend'][2020] = 300
        # hard_constraints['total_spend'][2030] = 400
        hard_constraints['total_spend'] = {}
        for t, progs in hard_constraints['programs'].items():
            # For every time at which programs are optimizable...
            if len(self.t) and t not in self.t:
                # If we are not wanting to constrain spending in this year, then
                # continue
                continue
            elif len(self.t):
                idx = np.where(self.t == t)[0][0]  # This is the index for the constraint year
            else:
                idx = None

            if not len(self.total_spend) or self.total_spend[idx] is None:
                # Get the total spend from the allocation in this year
                total_spend = 0.0
                for prog in progs:
                    total_spend += instructions.alloc[prog].get(t)
            else:
                total_spend = self.total_spend[idx]

            # Lastly, apply the budget factor
            if len(self.budget_factor) == 1:
                hard_constraints['total_spend'][t] = total_spend * self.budget_factor
            else:
                hard_constraints['total_spend'][t] = total_spend * self.budget_factor[idx]

        # Finally, for each adjustable, we need to store its upper and lower bounds
        # _in the year that the adjustment is being made_
        #
        # This is because in the case where we have an multiple adjustables targeting the same Program at different
        # time points, the upper/lower bounds might be different. For example, suppose we have spending on
        # Program 1 specified somehow as 2020=$20 and 2030=$30. Then suppose we add an Adjustment for Program 1 in
        # 2020 *and* 2030 with relative bounds of [0.5,2.0]. Then, in 2020 spending on Program 1 is limited to
        # $10-40 while in 2030 it's limited to $15-$60. Thus the spending constraint could also be time-varying.
        # In the case of a parametric overwrite, there will be no direct spending bounds based on the Adjustable. Then,
        # the spending should only be restricted to a positive value.

        hard_constraints['bounds'] = dict()

        for t, progs in hard_constraints['programs'].items():  # For each time point being constrained, and for each program
            if t not in hard_constraints['total_spend']:
                # If the time is not one where the total spending constraint is being applied, then
                # just skip it
                continue

            hard_constraints['bounds'][t] = dict()
            # If there is an Adjustable that reaches this Program in the appropriate year:

            # Keep track of the absolute lower and upper bounds on spending permitted by the program constraints
            minimum_spend = 0.0
            maximum_spend = 0.0

            for adjustment in optimization.adjustments:
                if hasattr(adjustment, 'prog_name') and adjustment.prog_name in progs and t in adjustment.t:
                    if isinstance(adjustment, SpendingAdjustment):
                        idx = np.where(adjustment.t == t)[0][0]  # If it is a SpendingAdjustment then set bounds from the appropriate Adjustable
                        adjustable = adjustment.adjustables[idx]
                        hard_constraints['bounds'][t][adjustment.prog_name] = adjustable.get_hard_bounds(instructions.alloc[adjustment.prog_name].get(t))  # The instructions should already have the initial spend on this program inserted. This may be inconsistent if multiple Adjustments reach the same program...!
                    else:
                        hard_constraints['bounds'][t][adjustment.prog_name] = (0.0, np.inf)  # If the Adjustment reaches spending but is not a SpendingAdjustment then do not constrain the alloc

                    minimum_spend += hard_constraints['bounds'][t][adjustment.prog_name][0]
                    maximum_spend += hard_constraints['bounds'][t][adjustment.prog_name][1]

            if minimum_spend > hard_constraints['total_spend'][t]:
                raise UnresolvableConstraint('The total spend in %.2f is constrained to %.2f but the individual programs have a total minimum spend of %.2f which is impossible to satisfy. Please either raise the total spending, or lower the minimum spend on one or more programs' % (t, hard_constraints['total_spend'][t], minimum_spend))

            if maximum_spend < hard_constraints['total_spend'][t]:
                raise UnresolvableConstraint('The total spend in %.2f is constrained to %.2f but the individual programs have a total maximum spend of %.2f which is impossible to satisfy. Please either lower the total spending, or raise the maximum spend on one or more programs' % (t, hard_constraints['total_spend'][t], maximum_spend))

        return hard_constraints

    def constrain_instructions(self, instructions, hard_constraints):

        penalty = 0.0

        for t, total_spend in hard_constraints['total_spend'].items():

            x0 = sc.odict()  # Order matters here
            bounds = []
            progs = hard_constraints['programs'][t]  # Programs eligible for constraining at this time
            LinearConstraint = [{'type': 'eq', 'fun': lambda x: np.sum(x) - total_spend, 'jac': lambda x: np.ones(x.shape)}]  # Constrain spend

            for prog in progs:
                x0[prog] = instructions.alloc[prog].get(t)
                bounds.append(hard_constraints['bounds'][t][prog])

            x0_array = np.array(x0.values()).ravel()
            x0_array_scaled = x0_array / sum(x0_array) * total_spend  # Multiplicative rescaling to match the total spend

            def jacfcn(x):
                dist = np.linalg.norm(x - x0_array_scaled)
                if dist == 0:
                    return np.zeros(x.shape)
                else:
                    return (x-x0_array_scaled)/dist

            res = scipy.optimize.minimize(lambda x: np.linalg.norm(x - x0_array_scaled), x0_array_scaled, jac=jacfcn, bounds=bounds, constraints=LinearConstraint, options={'maxiter': 500})

            if not res['success']:
                logger.error('TotalSpendConstraint failed - rejecting these proposed parameters with an objective value of np.inf')
                penalty = np.inf
            else:
                penalty += np.linalg.norm(res['x'] - x0_array)  # Penalty is the distance between the unconstrained budget and the constrained budget
                for name, val in zip(x0.keys(), res['x']):
                    instructions.alloc[name].insert(t, val)
        return penalty


class OptimInstructions(NamedItem):
    def __init__(self, json=None):
        self.name = json['name']
        self.json = json

    def make(self, project=None):
        proj = project
        name = self.json['name']
        parset_name = self.json['parset_name']  # WARNING, shouldn't be unused
        progset_name = self.json['progset_name']
        adjustment_year = self.json['start_year']  # The year when adjustments get made
        end_year = self.json['end_year']  # For cascades, this is the evaluation year. For other measurables, it is optimized from the adjustment year to the end year
        budget_factor = self.json['budget_factor']
        objective_weights = self.json['objective_weights']
        prog_spending = self.json['prog_spending']
        maxtime = self.json['maxtime']
        optim_type = self.json['optim_type']
        tool = self.json['tool']
        method = self.json.get('method', None)
        start_year = project.data.end_year  # The year when programs turn on

        if tool == 'cascade' and optim_type == 'money':
            raise NotImplementedError('Money minimization not yet implemented for Cascades tool')

        progset = proj.progsets[progset_name]  # Retrieve the progset

        # Set up the initial allocation and program instructions
        progset_instructions = ProgramInstructions(alloc=progset, start_year=start_year)  # passing in the progset means we fix the spending in the start year

        # Add a spending adjustment in the start/optimization year for every program in the progset, using the lower/upper bounds
        # passed in as arguments to this function
        adjustments = []
        default_spend = progset.get_alloc(tvec=adjustment_year, instructions=progset_instructions)  # Record the default spend for scale-up in money minimization
        for prog_name in progset.programs:
            limits = list(sc.dcp(prog_spending[prog_name]))
            if limits[0] is None:
                limits[0] = 0.0
            if limits[1] is None and optim_type == 'money':
                # Money minimization requires an absolute upper bound. Limit it to 5x default spend by default
                limits[1] = 5 * default_spend[prog_name]
            adjustments.append(SpendingAdjustment(prog_name, t=adjustment_year, limit_type='abs', lower=limits[0], upper=limits[1]))

            if optim_type == 'money':
                # Modify default spending to see if more money allows target to be met at all
                if limits[1] is not None and np.isfinite(limits[1]):
                    progset_instructions.alloc[prog_name].insert(adjustment_year, limits[1])
                else:
                    progset_instructions.alloc[prog_name] = TimeSeries(adjustment_year, 5 * default_spend[prog_name])

        if optim_type == 'outcome':
            # Add a total spending constraint with the given budget scale up
            # For money minimization we do not need to do this
            constraints = [TotalSpendConstraint(budget_factor=budget_factor)]
        else:
            constraints = None

        # Add all of the terms in the objective
        measurables = []
        for mname, mweight in objective_weights.items():

            if not mweight:
                continue

            if tool == 'cascade':
                tokens = mname.split(':')
                if tokens[0] == 'cascade_stage':  # Parse a measurable name like 'cascade_stage:Default:All diagnosed'
                    measurables.append(MaximizeCascadeStage(cascade_name=tokens[1], t=[end_year], pop_names='all', cascade_stage=tokens[2], weight=mweight))
                elif tokens[0] == 'conversion':  # Parse a measurable name like 'conversions:Default'
                    measurables.append(MaximizeCascadeConversionRate(cascade_name=tokens[1], t=[end_year], pop_names='all', weight=mweight))
                else:
                    raise Exception('Unknown measurable "%s"' % (mname))
            else:
                if optim_type == 'money':
                    # For money minimization, use at AtMostMeasurable to meet the target by the end year.
                    # The weight stores the threshold value
                    measurables.append(AtMostMeasurable(mname, t=end_year, threshold=mweight))
                else:
                    measurables.append(Measurable(mname, t=[adjustment_year, end_year], weight=mweight))

        if optim_type == 'money':
            # Do a prerun to convert the optimization targets into absolute units
            result = proj.run_sim(proj.parsets[parset_name], progset=progset, progset_instructions=progset_instructions, store_results=False)
            for measurable in measurables:
                val = measurable.get_objective_val(result.model)  # This is the baseline value for the quantity being thresholded
                assert measurable.threshold <= 100 and measurable.threshold >= 0
                measurable.threshold = val * (1 - measurable.threshold / 100.)

            # Then, add extra measurables for program spending
            for prog in progset.programs.values():
                measurables.append(MinimizeMeasurable(prog.name, adjustment_year))  # Minimize 2020 spending on Treatment 1

        # Create the Optimization object
        optim = Optimization(name=name, parsetname=parset_name, progsetname=progset_name, adjustments=adjustments, measurables=measurables, constraints=constraints, maxtime=maxtime)

        # Set the method used for optimization
        if method is not None:
            optim.method = method
        elif optim_type == 'money':
            optim.method = 'pso'

        return optim, progset_instructions


class Optimization(NamedItem):
    """
    Instructions on how to perform an optimization

    The Optimization object stores the information that defines an optimization operation.
    Optimization can be thought of as a function mapping one set of program instructions
    to another set of program instructions. The parameters of that function are stored in the
    Optimization object, and amount to

    - A definition of optimality
    - A specification of allowed changes to the program instructions
    - Any additional information required by a particular optimization algorithm e.g. ASD

    """

    def __init__(self, name=None, parsetname=None, progsetname=None, adjustments=None, measurables=None, constraints=None, maxtime=None, maxiters=None, method='asd'):

        # Get the name
        if name is None:
            name = 'default'
        NamedItem.__init__(self, name)

        self.parsetname = parsetname
        self.progsetname = progsetname
        self.maxiters = maxiters  # Not snake_case to match ASD
        self.maxtime = maxtime  # Not snake_case to match ASD
        self.method = method  # This gets passed to optimize() to select the algorithm

        assert adjustments is not None, 'Must specify some adjustments to carry out an optimization'
        assert measurables is not None, 'Must specify some measurables to carry out an optimization'
        self.adjustments = sc.promotetolist(adjustments)
        self.measurables = sc.promotetolist(measurables)

        if constraints:
            self.constraints = [constraints] if not isinstance(constraints, list) else constraints
        else:
            self.constraints = None

        if adjustments is None or measurables is None:
            raise Exception('Must supply either a json or an adjustments+measurables')
        return

    def __repr__(self):
        return sc.prepr(self)

    def get_initialization(self, progset, instructions):
        # Return arrays of lower and upper bounds for each adjustable
        x0 = []
        for adjustment in self.adjustments:
            x0 += adjustment.get_initialization(progset, instructions)
        x0 = np.array(x0)
        xmin = np.zeros(x0.shape)
        xmax = np.zeros(x0.shape)

        ptr = 0
        for adjustment in self.adjustments:
            for adjustable in adjustment.adjustables:
                bounds = adjustable.get_hard_bounds(x0[ptr])
                xmin[ptr] = bounds[0]
                xmax[ptr] = bounds[1]
                ptr += 1

        return x0, xmin, xmax

    def update_instructions(self, asd_values, instructions):
        idx = 0
        for adjustment in self.adjustments:
            adjustment.update_instructions(asd_values[idx:idx + len(adjustment.adjustables)], instructions)
            idx += len(adjustment.adjustables)

    def get_hard_constraints(self, x0, instructions):
        # Return hard constraints based on the starting initialization
        instructions = sc.dcp(instructions)
        self.update_instructions(x0, instructions)
        if not self.constraints:
            return None
        else:
            return [x.get_hard_constraint(self, instructions) for x in self.constraints]

    def constrain_instructions(self, instructions, hard_constraints):
        constraint_penalty = 0.0
        if self.constraints:
            for constraint, hard_constraint in zip(self.constraints, hard_constraints):
                constraint_penalty += constraint.constrain_instructions(instructions, hard_constraint)
        return constraint_penalty

    def compute_objective(self, model):
        # Take in a completed model run
        # Compute the objective based on the measurables
        # Program name will map to spending array for that program
        objective = 0.0
        for measurable in self.measurables:
            objective += measurable.eval(model)
        return objective


def _objective_fcn(asd_values, pickled_model=None, optimization=None, hard_constraints=None):
    # Compute the objective in ASD

    # Unpickle model
    model = pickle.loads(pickled_model)

    # Inject the ASD vector into the instructions
    optimization.update_instructions(asd_values, model.program_instructions)

    # Constrain the alloc - a penalty of `np.inf` signifies that the constraint could not be satisfied
    # In which case we can reject the proposed parameters _before_ processing the model
    constraint_penalty = optimization.constrain_instructions(model.program_instructions, hard_constraints)
    if np.isinf(constraint_penalty):
        return constraint_penalty

    # Use the updated instructions to run the model
    model.process()

    # Compute the objective function based on the model calculated values and return it
    # TODO - constraint_penalty is an interesting idea but need to make sure it doesn't dominate the actual objective
    # Using it directly seems to prioritize meeting the constraint and with ASD's single-parameter stepping this prevents convergence
    obj_val = 0.0 * constraint_penalty + optimization.compute_objective(model)

    return obj_val


def optimize(project, optimization, parset, progset, instructions, x0=None, xmin=None, xmax=None, hard_constraints=None):
    """
    Main user entry point for optimization

    The optional inputs `x0`, `xmin`, `xmax` and `hard_constraints` are used when
    performing parallel optimization (implementation not complete yet), in which case
    they are computed by the parallel wrapper to `optimize()`. Normally these variables
    would not be specified by users, because they are computed from the `Optimization`
    together with the instructions (because relative constraints in the Optimization are
    interpreted as being relative to the allocation in the instructions).

    :param project: A :class:`Project` instance
    :param optimization: An :class:`Optimization` instance
    :param parset: A :class:`ParameterSet` instance or name of a parset
    :param progset: A :class:`ProgramSet` instance or name of a progset
    :param instructions: A :class:`ProgramInstructions` instance
    :param x0: Not for manual use - override initial values
    :param xmin: Not for manual use - override lower bounds
    :param xmax: Not for manual use - override upper bounds
    :param hard_constraints: Not for manual use - override hard constraints
    :return: A :class:`ProgramInstructions` instance representing optimal instructions

    """
    # The ASD initialization, xmin and xmax values can optionally be
    # method can be one of
    # - asd (to use normal ASD)
    # - pso (to use particle swarm optimization from pyswarm)
    # - hyperopt (to use hyperopt's Bayesian optimization function)

    assert optimization.method in ['asd', 'pso', 'hyperopt']

    model = Model(project.settings, project.framework, parset, progset, instructions)
    pickled_model = pickle.dumps(model)

    initialization = optimization.get_initialization(progset, model.program_instructions)
    x0 = x0 if x0 is not None else initialization[0]
    xmin = xmin if xmin is not None else initialization[1]
    xmax = xmax if xmax is not None else initialization[2]

    if not hard_constraints:
        hard_constraints = optimization.get_hard_constraints(x0, model.program_instructions)  # The optimization passed in here knows how to calculate the hard constraints based on the program instructions

    # Prepare additional arguments for the objective function
    args = {
        'pickled_model': pickled_model,
        'optimization': optimization,
        'hard_constraints': hard_constraints,
    }

    # Check that the initial conditions are OK
    initial_objective = _objective_fcn(x0, **args)
    if not np.isfinite(initial_objective):
        raise InvalidInitialConditions('Optimization cannot begin because the objective function was NaN for the specified initialization')

    if optimization.method == 'asd':
        optim_args = {
            # 'stepsize': proj.settings.autofit_params['stepsize'],
            'maxiters': optimization.maxiters,
            'maxtime': optimization.maxtime,
            # 'sinc': proj.settings.autofit_params['sinc'],
            # 'sdec': proj.settings.autofit_params['sdec'],
            'fulloutput': False,
            'xmin': xmin,
            'xmax': xmax,
        }
        x_opt = sc.asd(_objective_fcn, x0, args, **optim_args)[0]
    elif optimization.method == 'pso':

        import pyswarm

        optim_args = {
            'maxiter': 3,
            'lb': xmin,
            'ub': xmax,
            'minstep': 1e-3,
            'debug': True
        }
        if np.any(~np.isfinite(xmin)) or np.any(~np.isfinite(xmax)):
            errormsg = 'PSO optimization requires finite upper and lower bounds to specify the search domain (i.e. every Adjustable needs to have finite bounds)'
            raise Exception(errormsg)

        x_opt, _ = pyswarm.pso(_objective_fcn, kwargs=args, **optim_args)
    elif optimization.method == 'hyperopt':

        import hyperopt
        import functools

        if np.any(~np.isfinite(xmin)) or np.any(~np.isfinite(xmax)):
            errormsg = 'hyperopt optimization requires finite upper and lower bounds to specify the search domain (i.e. every Adjustable needs to have finite bounds)'
            raise Exception(errormsg)

        space = []
        for i, (lower, upper) in enumerate(zip(xmin,xmax)):
            space.append(hyperopt.hp.uniform(str(i),lower,upper))
        fcn = functools.partial(_objective_fcn,**args) # Partial out the extra arguments to the objective

        optim_args = {
            'max_evals': optimization.maxiters if optimization.maxiters is not None else 100,
            'algo': hyperopt.tpe.suggest
        }

        x_opt = hyperopt.fmin(fcn, space, **optim_args)
        x_opt = np.array([x_opt[str(n)] for n in range(len(x_opt.keys()))])

    optimization.update_instructions(x_opt, model.program_instructions)
    optimization.constrain_instructions(model.program_instructions, hard_constraints)
    return model.program_instructions  # Return the modified instructions

# def parallel_optimize(project,optimization,parset,progset,program_instructions):
#
#    raise NotImplementedError
#
#    initialization = optimization.get_initialization(progset, program_instructions)
#    xmin = initialization[1]
#    xmax = initialization[2]
#
#    initial_instructions = program_instructions
#    for i in range(0,n_rounds):
#        for i in range(0,n_threads):
#            optimized_instructions[i] = optimize(optimization, parset, progset, initial_instructions, x0=None, xmin=xmin, xmax=xmax)
#        initial_instructions = pick_best(optimized_instructions)
#    return initial_instructions # Best one from the last round



"""
Define utility classes used throughout Atomica

"""

import inspect
import os
import ast
from bisect import bisect
import numpy as np
import scipy.interpolate
import sciris as sc
import itertools


def parent_dir():
    # Return the parent directory of the file that called this function
    return os.path.join(os.path.abspath(os.path.join(inspect.stack()[1][1], os.pardir)), '')


class NamedItem():
    def __init__(self, name:str=None):
        """
        NamedItem constructor

        A name must be a string

        :param name:
        """
        if name is None:
            name = '<unnamed>'
        self.name = name
        self.created = sc.now()
        self.modified = sc.now()

    def copy(self, name=None):
        x = sc.dcp(self)
        if name is not None:
            x.name = name
        return x

    def __repr__(self):
        return sc.prepr(self)


class NDict(sc.odict):
    """
    Store and sync items with a name property

    """

    def __init__(self, *args, **kwargs):
        sc.odict.__init__(self, *args, **kwargs)
        return None

    def __setitem__(self, key, item):
        # Store an item in the NDict with an explicitly provided key. If the item is a NamedItem
        # then the name of the item will be automatically synchronized - otherwise, it will just behave
        # like a normal odict
        sc.odict.__setitem__(self, key, item)

        # If it is a NamedItem, then synchronize the name of the object with the specified string key
        if sc.isstring(key) and isinstance(item, NamedItem):
            item.name = key
            item.modified = sc.now()
        return None

    def append(self, value):
        # Insert a NamedItem into the NDict by using the name of the item as the key. Of course this only
        # works for NamedItems, otherwise you have to explicitly provide the key
        if not isinstance(value, NamedItem):
            raise Exception('Can only automatically get the name from NamedItems. Instead of `x.append(y)` you need `x["name"]=y`')
        key = value.name
        sc.odict.append(self, key=key, value=value)
        return None

    def copy(self, old, new):
        """
        Copy an item and return the copy

        :param old: The key of the existing item
        :param new: The new name for the item
        :return: The copied item

        Example usage:
        >>> new_parset = proj.parsets.copy('old_name','new_name')

        """
        sc.odict.copy(self, old, new)
        if isinstance(self[new], NamedItem):
            self[new].name = new
        return self[new]


class TimeSeries(object):
    """
    Class to store time-series data

    Internally values are stored as lists rather than numpy arrays because
    insert/remove operations on lists tend to be faster (and working with sparse
    data is a key role of TimeSeries objects). Note that methods like :meth:`interpolate()`
    return numpy arrays, so the output types from such functions should generally match up
    with what is required by the calling function.

    :param t: Optionally specify a scalar, list, or array of time values
    :param vals: Optionally specify a scalar, list, or array of values (must be same size as ``t``)
    :param units: Optionally specify units (as a string)
    :param assumption: Optionally specify a scalar assumption
    :param sigma: Optionally specify a scalar uncertainty

    """

    def __init__(self, t=None, vals=None, units: str = None, assumption: float = None, sigma: float = None):
        t = sc.promotetolist(t) if t is not None else list()
        vals = sc.promotetolist(vals) if vals is not None else list()

        assert len(t) == len(vals)

        self.t = []  #: Sorted array of time points. Normally interacted with via methods like :meth:`insert()`
        self.vals = []  #: Time-specific values - indices correspond to ``self.t``
        self.units = units  #: The units of the quantity
        self.assumption = assumption  #: The time-independent scalar assumption
        self.sigma = sigma  #: Uncertainty value, assumed to be a standard deviation
        self._sampled = False  #: Flag to indicate whether sampling has been performed. Once sampling has been performed, cannot sample again

        for tx, vx in zip(t, vals):
            self.insert(tx, vx)

    def __repr__(self):
        output = sc.prepr(self)
        return output

    def __deepcopy__(self, memodict={}):
        new = TimeSeries.__new__(TimeSeries)
        new.t = self.t.copy()
        new.vals = self.vals.copy()
        new.units = self.units
        new.assumption = self.assumption
        new.sigma = self.sigma
        new._sampled = self._sampled
        return new

    # The operators for + and * below are prototypes but haven't been added
    # The reason is because they add a lot of complexity to the usage of the class
    # but it's not clear whether that complexity is worth it for the benefits it brings.
    # Part of the problem is also that propagation of uncertainty requires the time series
    # values which means that propagation of uncertainty is time dependent. So it's hard
    # to come up with intuitive behaviour for them.
    #
    # def __add__(self, other):
    #     """
    #     Add a constant or TimeSeries
    #
    #     If adding a constant, it is simply added onto the assumption and time values.
    #
    #     If adding another TimeSeries:
    #
    #         - The units must match
    #         - The assumption will be None if either TimeSeries has no assumption
    #         - The output times will be the union of times in both TimeSeries and they will
    #           be interpolated onto those times before adding. This preserves commutativity
    #           of addition
    #         - Uncertainties will be formally propagated
    #
    #     :param other: A TimeSeries or a scalar (float)
    #     :return: A new TimeSeries
    #
    #     """
    #
    #     if type(other) == type(self):
    #         # If we are operating on another TimeSeries
    #         if self.units is None:
    #             assert other.units is None, 'TimeSeries units must match for addition'
    #         else:
    #             assert self.units == other.units, 'TimeSeries units must match for addition'
    #
    #         new = self.copy()
    #
    #         # Add time values
    #         if self.t == other.t:
    #             new.t = [x+y for x,y in zip(self.vals, other.vals)]
    #         else:
    #             # Interpolate and add the time values
    #             t_out = sorted(set(self.t).union(set(other.t)))
    #             new.t = t_out
    #             new.vals = (self.interpolate(t_out) + other.interpolate(t_out)).tolist()
    #
    #         # Add the assumption
    #         if (self.assumption is None) or (other.assumption is None):
    #             new.assumption = None
    #         else:
    #             new.assumption = self.assumption + other.assumption
    #
    #         # Propagate uncertainty
    #         self_sigma = self.sigma if self.sigma else 0
    #         other_sigma = other.sigma if other.sigma else 0
    #
    #         if self_sigma or other_sigma:
    #             new.sigma = np.sqrt(self_sigma**2 + other_sigma**2)
    #     else:
    #         # Should be a numeric type
    #         new = self.copy()
    #         new.vals = [x+other for x in self.vals]
    #         new.assumption = self.assumption+other if self.assumption is not None else None
    #
    #     return new
    #
    # def __mul__(self, other):
    #     """
    #     Multiply by a constant or TimeSeries
    #
    #     If multiplying by a constant, it is simply operated on the assumption,
    #     values, and uncertainty.
    #
    #     If adding another TimeSeries:
    #
    #         - The units must match
    #         - The assumption will be None if either TimeSeries has no assumption
    #         - The output times will be the union of times in both TimeSeries and they will
    #           be interpolated onto those times before multiplying. This preserves commutativity
    #           of multiplication
    #         - Uncertainties will be formally propagated
    #
    #     :param other: A TimeSeries or a scalar (float)
    #     :return: A new TimeSeries
    #
    #     """
    #
    #     if type(other) == type(self):
    #         # If we are operating on another TimeSeries
    #         if self.units is None:
    #             assert other.units is None, 'TimeSeries units must match for addition'
    #         else:
    #             assert self.units == other.units, 'TimeSeries units must match for addition'
    #
    #         new = self.copy()
    #
    #         # Add time values
    #         if self.t == other.t:
    #             new.t = [x+y for x,y in zip(self.vals, other.vals)]
    #         else:
    #             # Interpolate and add the time values
    #             t_out = sorted(set(self.t).union(set(other.t)))
    #             new.t = t_out
    #             new.vals = (self.interpolate(t_out) * other.interpolate(t_out)).tolist()
    #
    #         # Add the assumption
    #         if (self.assumption is None) or (other.assumption is None):
    #             new.assumption = None
    #         else:
    #             new.assumption = self.assumption * other.assumption
    #
    #         # Propagate uncertainty
    #         if (self.sigma is None) or (other.sigma is None):
    #             new.sigma = None
    #         else:
    #             raise NotImplementedError # Proper propagation is time dependent
    #
    #     else:
    #         # Should be a numeric type
    #         new = self.copy()
    #         new.vals = [x*other for x in self.vals]
    #         new.assumption = self.assumption*other if self.assumption is not None else None
    #         new.assumption = self.sigma*other if self.sigma is not None else None
    #
    #     return new

    def copy(self):
        """
        Return a copy of the ``TimeSeries``

        :return: An independent copy of the ``TimeSeries``
        """

        return self.__deepcopy__(self)

    @property
    def has_data(self) -> bool:
        """
        Check if any data has been provided

        :return: ``True`` if any data has been entered (assumption or time-specific)

        """
        return self.assumption is not None or self.has_time_data

    @property
    def has_time_data(self) -> bool:
        """
        Check if time-specific data has been provided

        Unlike ``has_data``, this will return ``False`` if only an assumption has been entered

        :return: ``True`` if any time-specific data has been entered

        """
        # Returns true if any time-specific data has been entered (not just an assumption)
        return len(self.t) > 0

    def insert(self, t, v) -> None:
        """
        Insert a value at a particular time

        If the value already exists in the ``TimeSeries``, it will be overwritten/updated.
        The arrays are internally sorted by time value, and this order will be maintained.

        :param t: Time value to insert or update. If ``None``, the value will be assigned to the assumption
        :param v: Value to insert. If ``None``, this function will return immediately without doing anything

        """

        if v is None:  # Can't cast a None to a float, just skip it
            return

        v = float(v)  # Convert input to float

        if t is None:
            self.assumption = v
        elif t in self.t:
            idx = self.t.index(t)
            self.vals[idx] = v
        else:
            idx = bisect(self.t, t)
            self.t.insert(idx, t)
            self.vals.insert(idx, v)

    def get(self, t) -> float:
        """
        Retrieve value at a particular time

        This function will automatically retrieve the value of the assumption if
        no time specific values have been provided, or if any time specific values
        are provided, will return the value entered at that time. If time specific
        values have been entered and the requested time is not explicitly present,
        an error will be raised.

        This function may be deprecated in future because generally it is more useful
        to either call ``TimeSeries.interpolate()`` if interested in getting values at
        arbitrary times, or ``TimeSeries.get_arrays()`` if interested in retrieving
        values that have been entered.

        :param t: A time value. If ``None``, will return assumption regardless of whether
                  time data has been entered or not
        :return: The value at the corresponding time. Returns None if the value no value present
        """

        if t is None or len(self.t) == 0:
            return self.assumption
        elif t in self.t:
            return self.vals[self.t.index(t)]
        else:
            return None

    def get_arrays(self):
        """
        Return arrays with the contents of this TimeSeries

        The TimeSeries instance may have time values, or may simply have
        an assumption. If obtaining raw arrays is desired, this function will
        return arrays with values extracted from the appropriate attribute of the
        TimeSeries. However, in general, it is usually `.interpolate()` that is
        desired, rather than `.get_arrays()`

        :return: Tuple with two arrays - the first item is times (with a single NaN if
                 the TimeSeries only has an assumption) and the second item is values

        """
        if len(self.t) == 0:
            t = np.array([np.nan])
            v = np.array([self.assumption])
        else:
            t = np.array(self.t)
            v = np.array(self.vals)
        return t, v

    def remove(self, t):
        # To remove the assumption, set t=None
        if t is None:
            self.assumption = None
        elif t in self.t:
            idx = self.t.index(t)
            del self.t[idx]
            del self.vals[idx]
        else:
            raise Exception('Item not found')

    def remove_before(self,t_remove) -> None:
        """
        Remove times from start

        :param tval: Remove times up to but not including this time
        """

        for tval in sc.dcp(self.t):
            if tval < t_remove:
                self.remove(tval)

    def remove_after(self, t_remove) -> None:
        """
        Remove times from start

        :param tval: Remove times up to but not including this time
        """

        for tval in sc.dcp(self.t):
            if tval > t_remove:
                self.remove(tval)

    def remove_between(self, t_remove):
        # t is a two element vector [min,max] such that
        # times > min and < max are removed
        # Note that the endpoints are not included!
        for tval in sc.dcp(self.t):
            if t_remove[0] < tval < t_remove[1]:
                self.remove(tval)

    def interpolate(self, t2: np.array) -> np.array:
        """
        Return interpolated values

        :param t2: float, list, or array, with times
        :return: array the same length as t2, with interpolated values

        """

        t2 = sc.promotetoarray(t2)  # Deal with case where user prompts for single time point

        if not self.has_data:
            return np.full(t2.shape, np.nan)
        elif not self.has_time_data:
            return np.full(t2.shape, self.assumption)

        t1, v1 = self.get_arrays()
        return interpolate(t1, v1, t2)

    def sample(self, constant=True):
        """
        Return a sampled copy of the TimeSeries

        This method returns a copy of the TimeSeries in which the values have been
        perturbed based on the uncertainty value.

        :param constant: If True, time series will be perturbed by a single constant offset. If False,
                         an different perturbation will be applied to each time specific value independently.
        :return: A copied ``TimeSeries`` with perturbed values

        """

        if self._sampled:
            raise Exception('Sampling has already been performed - can only sample once')

        new = self.copy()
        if self.sigma is not None:
            delta = self.sigma * np.random.randn(1)[0]
            if self.assumption is not None:
                new.assumption += delta

            if constant:
                # Use the same delta for all data points
                new.vals = [v+delta for v in new.vals]
            else:
                # Sample again for each data point
                for i, (v, delta) in enumerate(zip(new.vals, self.sigma * np.random.randn(len(new.vals)))):
                    new.vals[i] = v+delta

        # Sampling flag only needs to be set if the TimeSeries had data to change
        if new.has_data:
            new._sampled = True

        return new


def evaluate_plot_string(plot_string: str):
    """
    Evaluate a plotting output specification

    The plots in the framework are specified as strings - for example,

    >>> plot_string = "{'New active DS-TB':['pd_div:flow','nd_div:flow']}"

    This needs to be (safely) evaluated so that the actual dict can be
    used. This function evaluates a string like this and returns a
    variable accordingly. For example

    >>> x = evaluate_plot_string("{'New active DS-TB':['pd_div:flow','nd_div:flow']}")

    is the same as

    >>> x = {'New active DS-TB':['pd_div:flow','nd_div:flow']}

    This will only happen if tokens associated with dicts and lists are present -
    otherwise the original string will just be returned directly

    :param plot_string: A string representation of Python structures (e.g., lists, dicts)
    :return: Evaluated expression, the same as if it has originally been entered in a .py file
    """

    if '{' in plot_string or '[' in plot_string:
        # Evaluate the string to set lists and dicts - do at least a little validation
        assert '__' not in plot_string, 'Cannot use double underscores in functions'
        assert len(plot_string) < 1800  # Function string must be less than 1800 characters
        fcn_ast = ast.parse(plot_string, mode='eval')
        for node in ast.walk(fcn_ast):
            if not (node is fcn_ast):
                assert isinstance(node, ast.Dict) or isinstance(node, ast.Str) or isinstance(node, ast.List) or isinstance(node, ast.Load), 'Only allowed to initialize lists and dicts of strings here'
        compiled_code = compile(fcn_ast, filename="<ast>", mode="eval")
        return eval(compiled_code)
    else:
        return plot_string


def format_duration(t: float, pluralize=False) -> str:
    """
    User-friendly string format of a duration

    This helper function is used when displaying durations in plots. It takes in
    a duration in units of years, and returns a string representation in user-friendly
    units. This function is mainly intended to be used to format denominators
    e.g., going from 'probability' and '1/365' to 'probability/day'

    :param t: A duration in units of years
    :param pluralize: Always return a plural suffix
    :return: A string

    Example usage:

    >>> format_duration(1)
    'year'
    >>> format_duration(1, pluralize=True)
    'years'
    >>> format_duration(1/365)
    'day'
    >>> format_duration(2/365)
    '2 days'
    >>> format_duration(1.5/52)
    '1.5 weeks'
    >>> format_duration(2/52)
    'fortnight'
    >>> format_duration(2.5/12)
    '2.5 months'

    """

    # First decide the base units
    if t >= 1.0:
        base_scale = 1
        timescale = 'year'
    elif t >= 1 / 12:
        base_scale = 1 / 12
        timescale = 'month'
    elif t >= 1 / 26:
        base_scale = 1 / 26
        timescale = 'fortnight'
    elif t >= 1 / 52:
        base_scale = 1 / 52
        timescale = 'week'
    else:
        base_scale = 1 / 365
        timescale = 'day'

    # Then, work out how many of the base unit there are
    converted_t = t / base_scale

    # If there is only one of the base unit, then return the timescale as the final string
    if abs(converted_t - 1.0) < 1e-5:
        return (timescale + 's') if pluralize else timescale
    elif converted_t % 1 < 1e-3:  # If it's sufficiently close to an integer, show it as an integer
        return '%d %ss' % (converted_t, timescale)
    else:
        return '%s %ss' % (sc.sigfig(converted_t, keepints=True, sigfigs=3), timescale)


def interpolate(x: np.array, y: np.array, x2: np.array, extrapolate=True) -> np.array:
    """
    pchip interpolation with constant extrapolation

    Atomica's standard interpolation routine is based on the pchip method. However, when
    extrapolation is desired (such as when interpolating parameters), it is most useful to
    assume the derivative is zero outside the original range of the function, rather than
    the default pchip behaviour of extrapolating with constant gradient.

    :param x: Original x values
    :param y: Original function values
    :param x2: New desired x values
    :param extrapolate: If True, use constant interpolation outside the original domain. Otherwise,
                        the function value will be NaN. Note that in general, model outputs
                        should not be extrapolated
    :return: Array the same size as ``x2`` with interpolated values

    """

    # Remove NaNs - note that advanced indexing means that the variable is copied
    idx = ~np.isnan(x) & ~np.isnan(y)
    x = x[idx]
    y = y[idx]

    if x.size == 0:
        raise Exception('No time points remained after removing NaNs from the TimeSeries')
    elif x.size == 1:
        return np.full(x2.shape, y[0])
    else:
        f = scipy.interpolate.PchipInterpolator(x, y, axis=0, extrapolate=False)
        if extrapolate:
            y2 = np.zeros(x2.shape)
            y2[(x2 >= x[0]) & (x2 <= x[-1])] = f(x2[(x2 >= x[0]) & (x2 <= x[-1])])
            y2[x2 < x[0]] = y[0]
            y2[x2 > x[-1]] = y[-1]
        else:
            y2 = f(x2)  # PchipInterpolator will return NaNs outside the domain
        return y2


def floor_interpolator(x, y):
    """
    Stepped (single-sided nearest neighbour) interpolation

    This returns a function that does interpolation where the return value
    corresponds to the nearest smaller neighbour (and NaN if extrapolating)

    :param x: Original x values
    :param y: Original y values
    :return: Function object `f(x)` that performs interpolation

    Example usage:

    >>> f = floor_interpolator([1,2,3,4],[1,2,3,4])
    >>> (f(0.5) )
    [nan]
    >>> (f(1))
    [1.]
    >>> (f(1.5))
    [1.]
    >>> (f(2))
    [2.]
    >>> (f(4))
    [4.]
    >>> (f(5))
    [nan]
    >>> (f([0.5,1,1.5,2,4,5]))
    [nan  1.  1.  2.  4. nan]

    """
    # Return a function that returns the floor interpolation for given values (and NaN if out of range)
    x = sc.promotetoarray(x)
    y = sc.promotetoarray(y)

    idx = np.argsort(x)
    x = x[idx]
    y = y[idx]

    def f(x2):
        x2 = sc.promotetoarray(x2)
        out = np.full(x2.shape, np.nan)
        for i, v in enumerate(x2):
            if v > x[-1]:
                continue

            flt = np.where(x <= v)[0]
            if not len(flt):
                continue

            out[i] = y[flt[-1]]

        return out

    return f


def nested_loop(inputs, loop_order):
    """
    Zip list of lists in order

    This is used in :func:`plot_bars` to control whether 'times' or 'results' are the
    outer grouping. This function takes in a list of lists to iterate over, and their
    nesting order. It then yields tuples of items in the given order. Only tested
    for two levels (which are all that get used in :func:`plot_bars` but in theory
    supports an arbitrary number of items.

    :param inputs: List of lists. All lists should have the same length
    :param loop_order: Nesting order for the lists
    :return: Generator yielding tuples of items, one for each list

    Example usage:

    >>> list(nested_loop([['a','b'],[1,2]],[0,1]))
    [['a', 1], ['a', 2], ['b', 1], ['b', 2]]

    Notice how the first two items have the same value for the first list
    while the items from the second list vary. If the `loop_order` is
    reversed, then:

    >>> list(nested_loop([['a','b'],[1,2]],[1,0]))
    [['a', 1], ['b', 1], ['a', 2], ['b', 2]]

    Notice now how now the first two items have different values from the
    first list but the same items from the second list.

    """

    loop_order = list(loop_order) # Convert to list, in case loop order was passed in as a generator e.g. from map()
    inputs = [inputs[i] for i in loop_order]
    iterator = itertools.product(*inputs)  # This is in the loop order
    for item in iterator:
        out = [None] * len(loop_order)
        for i in range(len(item)):
            out[loop_order[i]] = item[i]
        yield out
